/* 
 * Copyright 2008 Maurice Castro. All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *    1. Redistributions of source code must retain the above copyright
 *        notice, this list of conditions and the following disclaimer.
 *    2. Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials provided
 *        with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY Maurice Castro ``AS IS'' AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL Maurice Castro OR OTHER CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
 * OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
 * OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 * BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <usbhid.h>
#include <dev/usb/usb.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <string.h>
#include <math.h>
#include <errno.h>
#include "k8055.h"

input_t hid_read_input(k8055_dev_t *d);
int hid_write_output(k8055_dev_t *d, output_t out);
input_t gen_read_input(k8055_dev_t *d);
int gen_write_output(k8055_dev_t *d, output_t out);

/*
 * this code is based on the documentation found in the libk8055 library 
 * developed by Sven Lindberg and Pjetur G. Hjaltason. It is a complete
 * reimplementation of their work and does not use any of their code. Their
 * code can be found at http://libk8055.sourceforge.net/
 *
 * the debounce timer setting algorithm is based on the algorithm described
 * in their internal documentation
 */

#define RETRY 3
#define USBDEV "/dev/usb"
#define DEVDIR "/dev/"
#define MAX_USB_BUS 10

#define VELLEMAN_VENDOR_ID 0x10cf
#define K8055_PID 0x5500

extern int debug;
extern int errno;

/* ************************************************************************* */
/* locate velleman card device                                               */
/* ************************************************************************* */

static char *device(int fd, int d, int card_no)
{
	int e;
	int i;
	struct usb_device_info di;
	char *r;
	di.udi_addr = d;
	e = ioctl(fd, USB_DEVICEINFO, &di);
        if (e) 
	{
		if (debug & 0x1)
			if (errno != ENXIO)
				perror("IOCTL Failed");
                return NULL;
        }
	/* if the vendor and product numbers match we report the device */
	/* it is connected to */
	if (debug & 0x1)
		fprintf(stderr,"Found Vendor %x Product %x\n", di.udi_vendorNo, di.udi_productNo);
	if (di.udi_vendorNo == VELLEMAN_VENDOR_ID && di.udi_productNo == K8055_PID + card_no)
	{
		if (debug & 0x1)
			fprintf(stderr,"Found card\n");
		for (i = 0; i < USB_MAX_DEVNAMES; i++)
		{
			if (di.udi_devnames[i][0])
			{
				if (debug & 0x1)
					fprintf(stderr,"Device %s\n", di.udi_devnames[i]);
				r = malloc(strlen(DEVDIR) + strlen(di.udi_devnames[i]) + 1);
				if (r == NULL)
				{
					if (debug)
						fprintf(stderr, "Unable to allocate memory for device name\n");
					return NULL;
				}
				strcpy(r, DEVDIR);
				strcat(r, di.udi_devnames[i]);
				return(r);
			}
		}
	}
	return NULL;
}

static char *bus(int fd, int card_no)
{
	int i;
	char *r;
	for (i = 1; i < USB_MAX_DEVICES; i++)
	{
		r = device(fd, i, card_no);
		if (r)
			return(r);
	}
	return NULL;
}

char *find_k8055(int card_no)
{
	int i;
	char *r;
	int fd;
	char buf[50];

	for (i=0; i < MAX_USB_BUS; i++)
	{
		snprintf(buf, sizeof(buf), "%s%d", USBDEV, i);
		fd = open(buf, O_RDONLY);
		if (fd >= 0) 
		{
			r = bus(fd, card_no);
			if (r)
				return(r);
			close(fd);
		}
		else
		{
			if (errno == ENOENT || errno == ENXIO)
				continue;
			if (debug & 0x1)
				perror(buf);
		}
	}
	return NULL;
}

/* ************************************************************************* */
/* device test                                                               */
/* ************************************************************************* */

int is_hid_dev(char *dev)
{
	int report_id;
	int fd;

	fd = open(dev, O_RDWR);
	if (fd == -1)
	{
		return -1;
	}
	/* report id */
	if (ioctl(fd, USB_GET_REPORT_ID, &report_id) < 0)
	{
		if (errno == EINVAL)
		{
			close(fd);
			return 0;
		}
	}
	close(fd);
	return(1);
}

/* ************************************************************************* */
/* access functions                                                          */
/* ************************************************************************* */

/* generic read function */

input_t read_input(k8055_dev_t *d)
{
	return(d->read_input(d));
}

/* close the device
 */

void close_usb(k8055_dev_t *d)
{
	if (d->fd)
		close(d->fd);
	if (d->fd_ep1)
		close(d->fd_ep1);
}

/* open target usb hid device
 *
 * open the device we are passed and verify that it has the right read
 * and write sizes for a k8055 - both are 8 bytes in size
 * return the file descriptor
 */

k8055_dev_t *open_usbhid(char *dev)
{
	report_desc_t rd;
	int report_id;
	k8055_dev_t *d;

	d = malloc(sizeof(*d));
	if (!d)
		return NULL;

	/* only required if using display functions */
	hid_init(NULL);
	d->fd = open(dev, O_RDWR);
	if (d->fd == -1)
	{
		if (debug)
			fprintf(stderr, "%s\n", dev);
		return NULL;
	}
	/* report id */
	if (ioctl(d->fd, USB_GET_REPORT_ID, &report_id) < 0)
	{
		if (errno == EINVAL)
		{
			/* we are probably not a hid device */
			if (debug)
				fprintf(stderr, "Failed to get HID Report\n");
			return NULL;
		}
                report_id = -1;
	}
	if (debug & 0x2)
	{
		fprintf(stderr, "Report ID: %d\n", report_id);
	}

	/* get the report descriptor and test for match to target */
	rd = hid_get_report_desc(d->fd);
	if (rd == NULL)
	{
		if (debug)
			fprintf(stderr, "Failed to get report descriptor\n");
		return NULL;
	}
	if (hid_report_size(rd, hid_input, report_id) != 8)
	{
		if (debug)
			fprintf(stderr, "Invalid read block size\n");
		return NULL;
	}
	if (hid_report_size(rd, hid_output, report_id) != 8)
	{
		if (debug)
			fprintf(stderr, "Invalid write block size\n");
		return NULL;
	}
	d->write_output = hid_write_output;
	d->read_input = hid_read_input;
	return(d);
}

/* read the 8 bytes that are sent by the card into the input buffer */
/* we try a maximum of RETRY times for a read to work then abort */
/* this has not proved necessary in practice */
input_t hid_read_input(k8055_dev_t *d)
{
	input_t in;
	int retry;
	in.err = 0;
	for (retry = 0; retry < RETRY; retry++)
	{
		if (sizeof(in.b) == read(d->fd, &in.b, sizeof(in.b)))
		{
			if (debug & 0x4)
			{
				int i;
				fprintf(stderr, "Received: ");
				for  (i = 0; i < sizeof(in.b); i++)
					fprintf(stderr, "%02x ", in.b[i]);
				fprintf(stderr, "\n");
			}
			return in;
		}
		if (debug & 0x4)
		{
			fprintf(stderr, "Read retry %d failed\n", retry);
		}
	}
	if (debug)
		fprintf(stderr, "Read failure\n");
	in.err = 1;
	return in;
}

/* write the 8 bytes that are sent by the card */
/* we try a maximum of RETRY times for a write to work then abort */
/* this has not proved necessary in practice */
int hid_write_output(k8055_dev_t *d, output_t out)
{
	int retry;
	if (debug & 0x4)
	{
		int i;
		fprintf(stderr, "Sending: ");
		for  (i = 0; i < sizeof(out.b); i++)
			fprintf(stderr, "%02x ", out.b[i]);
		fprintf(stderr, "\n");
	}
	for (retry = 0; retry < RETRY; retry++)
	{
		if (sizeof(out.b) == write(d->fd, &out.b, sizeof(out.b)))
			return 0;
		if (debug & 0x4)
		{
			fprintf(stderr, "Write retry %d failed\n", retry);
		}
	}
	if (debug)
		fprintf(stderr, "Write failure\n");
	return -1;
}

/* open target usb ugen device
 *
 * open the device we are passed 
 */

#define EP1 ".1"

k8055_dev_t *open_usbgen(char *dev)
{
	k8055_dev_t *d;
	char *ep1;

	d = malloc(sizeof(*d));
	if (!d)
		return NULL;

	/* only required if using display functions */
	d->fd = open(dev, O_RDWR);
	if (d->fd == -1)
	{
		if (debug)
			fprintf(stderr, "%s\n", dev);
		return NULL;
	}
	ep1 = malloc(strlen(dev) + strlen(EP1)+1);
	if (!ep1)
	{
		free(d);
		return NULL;
	}
	ep1 = strdup(dev);
	strcat(ep1, EP1);
	d->fd_ep1 = open(ep1, O_RDWR);
	if (d->fd_ep1 == -1)
	{
		if (debug)
			fprintf(stderr, "%s\n", ep1);
		return NULL;
	}
	d->write_output = gen_write_output;
	d->read_input = gen_read_input;
	return(d);
}

/* read the 8 bytes that are sent by the card into the input buffer */
/* we try a maximum of RETRY times for a read to work then abort */
/* this has not proved necessary in practice */
input_t gen_read_input(k8055_dev_t *d)
{
	input_t in;
	int retry;
	int one = 1;
	int timeout = 1000; /* 1000ms */
	in.err = 0;
	if (ioctl(d->fd_ep1, USB_SET_TIMEOUT, &timeout) < 0)
	{
		if (debug)
			fprintf(stderr, "Unable to set timeout\n");
		in.err = -1;
		return in;
	}
	if (ioctl(d->fd_ep1, USB_SET_SHORT_XFER, &one) < 0)
	{
		if (debug)
			fprintf(stderr, "Unable to set short transfer\n");
		in.err = -1;
		return in;
	}

	for (retry = 0; retry < RETRY; retry++)
	{
		if (sizeof(in.b) == read(d->fd_ep1, &in.b, sizeof(in.b)))
		{
			if (debug & 0x4)
			{
				int i;
				fprintf(stderr, "Received: ");
				for  (i = 0; i < sizeof(in.b); i++)
					fprintf(stderr, "%02x ", in.b[i]);
				fprintf(stderr, "\n");
			}
			return in;
		}
		if (debug & 0x4)
		{
			fprintf(stderr, "Read retry %d failed\n", retry);
		}
	}
	if (debug)
		fprintf(stderr, "Read failure\n");
	in.err = -1;
	return in;
}

/* write the 8 bytes that are sent by the card */
/* we try a maximum of RETRY times for a write to work then abort */
/* this has not proved necessary in practice */
int gen_write_output(k8055_dev_t *d, output_t out)
{
	int retry;
	int timeout = 1000; /* 1000ms */
	if (ioctl(d->fd_ep1, USB_SET_TIMEOUT, &timeout) < 0)
	{
		if (debug)
			fprintf(stderr, "Unable to set timeout\n");
		return -1;
	}
	if (debug & 0x4)
	{
		int i;
		fprintf(stderr, "Sending: ");
		for  (i = 0; i < sizeof(out.b); i++)
			fprintf(stderr, "%02x ", out.b[i]);
		fprintf(stderr, "\n");
	}
	for (retry = 0; retry < RETRY; retry++)
	{
		if (sizeof(out.b) == write(d->fd_ep1, &out.b, sizeof(out.b)))
			return 0;
		if (debug & 0x4)
		{
			fprintf(stderr, "Write retry %d failed\n", retry);
		}
	}
	if (debug)
		fprintf(stderr, "Write failure\n");
	return -1;
}

/* ************************************************************************* */
/* input decoding "functions"                                                */
/* ************************************************************************* */

/* 16 bit counter 1 */
uint16_t counter1(input_t in)
{
	uint16_t r;
	r = in.b[4];
	r += in.b[5] << 8;
	return r;
}

/* 16 bit counter 2 */
uint16_t counter2(input_t in)
{
	uint16_t r;
	r = in.b[6];
	r += in.b[7] << 8;
	return r;
}

/* ************************************************************************* */
/* output functions                                                          */
/* ************************************************************************* */

/* called before reading or writing to device after openning */
int reset(k8055_dev_t *d)
{
	output_t out;
	memset(&out.b, 0x0, sizeof(out.b));
	return(d->write_output(d, out));
}

int reset_counter1(k8055_dev_t *d)
{
	output_t out;
	out.b[0] = 0x3;
	out.b[4] = 0x0;
	return(d->write_output(d, out));
}

int reset_counter2(k8055_dev_t *d)
{
	output_t out;
	out.b[0] = 0x4;
	out.b[5] = 0x0;
	return(d->write_output(d, out));
}

/* set analog and digital outputs to arguments given, the digital output is */
/* a bitmap */
int set_anadig(k8055_dev_t *d, uint8_t dig, uint8_t an1, uint8_t an2)
{
	output_t out;
	out.b[0] = 0x5;
	out.b[1] = dig;
	out.b[2] = an1;
	out.b[3] = an2;
	return(d->write_output(d, out));
}

/* set the debounce time, in milliseconds, for the counters */
int set_debounce1(k8055_dev_t *d, long ms)
{
	output_t out;
	float t;

	t = ms;
	if (t > 7450.0)
		t = 7450.0;
	t = sqrtf(t / 0.115);
	out.b[0] = 0x1;
	out.b[6] = (t + 0.5);
	return(d->write_output(d, out));
}

int set_debounce2(k8055_dev_t *d, long ms)
{
	output_t out;
	float t;

	t = ms;
	if (t > 7450.0)
		t = 7450.0;
	t = sqrtf(t / 0.115);
	out.b[0] = 0x2;
	out.b[7] = (t + 0.5);
	return(d->write_output(d, out));
}
